@@ -1,5 +1,5 @@
 /* evilwm - Minimalist Window Manager for X
- * Copyright (C) 1999-2007 Ciaran Anscomb <evilwm@6809.org.uk>
+ * Copyright (C) 1999-2011 Ciaran Anscomb <evilwm@6809.org.uk>
  * see README for license and other details. */
 
 #include <stdio.h>
@@ -7,20 +7,65 @@
 #include "evilwm.h"
 #include "log.h"
 
+#define MAXIMUM_PROPERTY_LENGTH 4096
+
 static int send_xmessage(Window w, Atom a, long x);
 
 /* used all over the place.  return the client that has specified window as
  * either window or parent */
 
 Client *find_client(Window w) {
-	Client *c;
+	struct list *iter;
 
-	for (c = head_client; c; c = c->next)
+	for (iter = clients_tab_order; iter; iter = iter->next) {
+		Client *c = iter->data;
 		if (w == c->parent || w == c->window)
 			return c;
+	}
 	return NULL;
 }
 
+void client_hide(Client *c) {
+	c->ignore_unmap++;  /* Ignore unmap so we don't remove client */
+	XUnmapWindow(dpy, c->parent);
+	set_wm_state(c, IconicState);
+}
+
+void client_show(Client *c) {
+	XMapWindow(dpy, c->parent);
+	set_wm_state(c, NormalState);
+}
+
+void client_raise(Client *c) {
+	XRaiseWindow(dpy, c->parent);
+	clients_stacking_order = list_to_tail(clients_stacking_order, c);
+	ewmh_set_net_client_list_stacking(c->screen);
+}
+
+/* This doesn't just call XLowerWindow(), as that could push the window
+ * below "DESKTOP" type windows we're not managing. */
+void client_lower(Client *c) {
+	struct list *iter;
+	Client *below;
+	Window order[2];
+	/* Find lowest other client in stacking order that is visible on the
+	 * same screen. */
+	for (iter = clients_stacking_order; iter; iter = iter->next) {
+		below = iter->data;
+		if (below == c)
+			return;
+		if (below->screen == c->screen && (is_fixed(below) || below->vdesk == c->screen->vdesk))
+			break;
+	}
+	if (!iter) return;
+	order[0] = below->parent;
+	order[1] = c->parent;
+	XRestackWindows(dpy, order, 2);
+	clients_stacking_order = list_delete(clients_stacking_order, c);
+	clients_stacking_order = list_insert_before(clients_stacking_order, iter, c);
+	ewmh_set_net_client_list_stacking(c->screen);
+}
+
 void set_wm_state(Client *c, int state) {
 	/* Using "long" for the type of "data" looks wrong, but the
 	 * fine people in the X Consortium defined it this way
@@ -50,46 +95,47 @@ void send_config(Client *c) {
 	XSendEvent(dpy, c->window, False, StructureNotifyMask, (XEvent *)&ce);
 }
 
-/* Support for 'gravitating' clients based on their original
- * border width and configured window manager frame width. */
-void gravitate_client(Client *c, int sign) {
-	int d0 = sign * c->border;
-	int d1 = sign * c->old_border;
-	int d2 = sign * (2*c->old_border - c->border);
+/* Shift client to show border according to window's gravity. */
+void gravitate_border(Client *c, int bw) {
+	int dx = 0, dy = 0;
 	switch (c->win_gravity) {
-		case NorthGravity:
-			c->x += d1;
-			c->y += d0;
-			break;
-		case NorthEastGravity:
-			c->x += d2;
-			c->y += d0;
-			break;
-		case EastGravity:
-			c->x += d2;
-			c->y += d1;
-			break;
-		case SouthEastGravity:
-			c->x += d2;
-			c->y += d2;
-			break;
-		case SouthGravity:
-			c->x += d1;
-			c->y += d2;
-			break;
-		case SouthWestGravity:
-			c->x += d0;
-			c->y += d2;
-			break;
-		case WestGravity:
-			c->x += d0;
-			c->y += d1;
-			break;
-		case NorthWestGravity:
-		default:
-			c->x += d0;
-			c->y += d0;
-			break;
+	default:
+	case NorthWestGravity:
+		dx = bw;
+		dy = bw;
+		break;
+	case NorthGravity:
+		dy = bw;
+		break;
+	case NorthEastGravity:
+		dx = -bw;
+		dy = bw;
+		break;
+	case EastGravity:
+		dx = -bw;
+		break;
+	case CenterGravity:
+		break;
+	case WestGravity:
+		dx = bw;
+		break;
+	case SouthWestGravity:
+		dx = bw;
+		dy = -bw;
+		break;
+	case SouthGravity:
+		dy = -bw;
+		break;
+	case SouthEastGravity:
+		dx = -bw;
+		dy = -bw;
+		break;
+	}
+	if (c->x != 0 || c->width != DisplayWidth(dpy, c->screen->screen)) {
+		c->x += dx;
+	}
+	if (c->y != 0 || c->height != DisplayHeight(dpy, c->screen->screen)) {
+		c->y += dy;
 	}
 }
 
@@ -99,32 +145,36 @@ void select_client(Client *c) {
 	if (c) {
 		unsigned long bpixel;
 #ifdef VWM
-		if (is_sticky(c))
+		if (is_fixed(c))
 			bpixel = c->screen->fc.pixel;
 		else
 #endif
 			bpixel = c->screen->fg.pixel;
 		XSetWindowBorder(dpy, c->parent, bpixel);
-#ifdef COLOURMAP
 		XInstallColormap(dpy, c->cmap);
-#endif
 		XSetInputFocus(dpy, c->window, RevertToPointerRoot, CurrentTime);
 	}
 	current = c;
+	ewmh_set_net_active_window(c);
 }
 
 #ifdef VWM
-void fix_client(Client *c) {
-	toggle_sticky(c);
-	select_client(c);
-	update_net_wm_state(c);
+void client_to_vdesk(Client *c, unsigned int vdesk) {
+	if (valid_vdesk(vdesk)) {
+		c->vdesk = vdesk;
+		if (c->vdesk == c->screen->vdesk || c->vdesk == VDESK_FIXED) {
+			client_show(c);
+		} else {
+			client_hide(c);
+		}
+		ewmh_set_net_wm_desktop(c);
+		select_client(current);
+	}
 }
 #endif
 
 void remove_client(Client *c) {
-	Client *p;
-
-	LOG_DEBUG("remove_client() : Removing...\n");
+	LOG_ENTER("remove_client(window=%lx, %s)", c->window, c->remove ? "withdrawing" : "wm quitting");
 
 	XGrabServer(dpy);
 	ignore_xerror = 1;
@@ -139,42 +189,50 @@ void remove_client(Client *c) {
 	 *  place when it is shutting down." (both _NET_WM_DESKTOP and
 	 *  _NET_WM_STATE) */
 	if (c->remove) {
-		LOG_DEBUG("\tremove_client() : setting WithdrawnState\n");
+		LOG_DEBUG("setting WithdrawnState\n");
 		set_wm_state(c, WithdrawnState);
-#ifdef VWM
-		XDeleteProperty(dpy, c->window, xa_net_wm_desktop);
-		XDeleteProperty(dpy, c->window, xa_net_wm_state);
-#endif
+		ewmh_withdraw_client(c);
+	} else {
+		ewmh_deinit_client(c);
 	}
 
-	ungravitate(c);
+	gravitate_border(c, -c->border);
+	gravitate_border(c, c->old_border);
+	c->x -= c->old_border;
+	c->y -= c->old_border;
 	XReparentWindow(dpy, c->window, c->screen->root, c->x, c->y);
 	XSetWindowBorderWidth(dpy, c->window, c->old_border);
 	XRemoveFromSaveSet(dpy, c->window);
 	if (c->parent)
 		XDestroyWindow(dpy, c->parent);
 
-	if (head_client == c) head_client = c->next;
-	else for (p = head_client; p && p->next; p = p->next)
-		if (p->next == c) p->next = c->next;
+	clients_tab_order = list_delete(clients_tab_order, c);
+	clients_mapping_order = list_delete(clients_mapping_order, c);
+	clients_stacking_order = list_delete(clients_stacking_order, c);
+	/* If the wm is quitting, we'll remove the client list properties
+	 * soon enough, otherwise: */
+	if (c->remove) {
+		ewmh_set_net_client_list(c->screen);
+		ewmh_set_net_client_list_stacking(c->screen);
+	}
 
 	if (current == c)
 		current = NULL;  /* an enter event should set this up again */
 	free(c);
 #ifdef DEBUG
 	{
-		Client *pp;
+		struct list *iter;
 		int i = 0;
-		for (pp = head_client; pp; pp = pp->next)
+		for (iter = clients_tab_order; iter; iter = iter->next)
 			i++;
-		LOG_DEBUG("\tremove_client() : free(), window count now %d\n", i);
+		LOG_DEBUG("free(), window count now %d\n", i);
 	}
 #endif
 
 	XUngrabServer(dpy);
 	XSync(dpy, False);
 	ignore_xerror = 0;
-	LOG_DEBUG("remove_client() returning\n");
+	LOG_LEAVE();
 }
 
 void send_wm_delete(Client *c, int kill_client) {
@@ -223,3 +281,20 @@ void set_shape(Client *c) {
 	}
 }
 #endif
+
+void *get_property(Window w, Atom property, Atom req_type,
+	           unsigned long *nitems_return) {
+	Atom actual_type;
+	int actual_format;
+	unsigned long bytes_after;
+	unsigned char *prop;
+	if (XGetWindowProperty(dpy, w, property,
+	                       0L, MAXIMUM_PROPERTY_LENGTH / 4, False,
+	                       req_type, &actual_type, &actual_format,
+	                       nitems_return, &bytes_after, &prop) == Success) {
+		if (actual_type == req_type)
+			return (void *)prop;
+		XFree(prop);
+	}
+	return NULL;
+}
