@@ -1,5 +1,5 @@
 /* evilwm - Minimalist Window Manager for X
- * Copyright (C) 1999-2007 Ciaran Anscomb <evilwm@6809.org.uk>
+ * Copyright (C) 1999-2011 Ciaran Anscomb <evilwm@6809.org.uk>
  * see README for license and other details. */
 
 #include <X11/X.h>
@@ -20,6 +20,7 @@
 #endif
 
 #include "keymap.h"
+#include "list.h"
 
 /* Required for interpreting MWM hints: */
 #define _XA_MWM_HINTS           "_MOTIF_WM_HINTS"
@@ -54,14 +55,27 @@ typedef struct {
 
 /* readability stuff */
 
+#define VDESK_NONE  (0xfffffffe)
+#define VDESK_FIXED (0xffffffff)
+#define VDESK_MAX   (7)
 #define KEY_TO_VDESK(key) ((key) - XK_1)
-#define valid_vdesk(v) ((unsigned)(v) < KEY_TO_VDESK(XK_8))
+#define valid_vdesk(v) ((v) == VDESK_FIXED || (v) <= VDESK_MAX)
 
 #define RAISE           1
 #define NO_RAISE        0       /* for unhide() */
 
-#define MAXIMISE_HORZ        (1<<0)
-#define MAXIMISE_VERT        (1<<1)
+/* EWMH hints use these definitions, so for simplicity my functions
+ * will too: */
+#define NET_WM_STATE_REMOVE     0    /* remove/unset property */
+#define NET_WM_STATE_ADD        1    /* add/set property */
+#define NET_WM_STATE_TOGGLE     2    /* toggle property  */
+
+/* EWMH window type bits */
+#define EWMH_WINDOW_TYPE_DESKTOP (1<<0)
+#define EWMH_WINDOW_TYPE_DOCK    (1<<1)
+
+#define MAXIMISE_HORZ   (1<<0)
+#define MAXIMISE_VERT   (1<<1)
 
 /* some coding shorthand */
 
@@ -74,15 +88,15 @@ typedef struct {
 	None, curs, CurrentTime) == GrabSuccess)
 #define grab_button(w, mask, button) do { \
 		XGrabButton(dpy, button, (mask), w, False, ButtonMask, \
-			    GrabModeAsync, GrabModeSync, None, None); \
+		            GrabModeAsync, GrabModeSync, None, None); \
 		XGrabButton(dpy, button, LockMask|(mask), w, False, ButtonMask,\
-			    GrabModeAsync, GrabModeSync, None, None); \
+		            GrabModeAsync, GrabModeSync, None, None); \
 		XGrabButton(dpy, button, numlockmask|(mask), w, False, \
-			    ButtonMask, GrabModeAsync, GrabModeSync, \
-			    None, None); \
+		            ButtonMask, GrabModeAsync, GrabModeSync, \
+		            None, None); \
 		XGrabButton(dpy, button, numlockmask|LockMask|(mask), w, False,\
-			    ButtonMask, GrabModeAsync, GrabModeSync, \
-			    None, None); \
+		            ButtonMask, GrabModeAsync, GrabModeSync, \
+		            None, None); \
 	} while (0)
 #define setmouse(w, x, y) XWarpPointer(dpy, None, w, 0, 0, 0, 0, x, y)
 #define get_mouse_position(xp,yp,root) do { \
@@ -91,19 +105,10 @@ typedef struct {
 		unsigned int dui; \
 		XQueryPointer(dpy, root, &dw, &dw, xp, yp, &di, &di, &dui); \
 	} while (0)
-#define gravitate(c) gravitate_client(c, 1)
-#define ungravitate(c) gravitate_client(c, -1)
 
-#define is_sticky(c) (c->sticky)
-#define add_sticky(c) c->sticky = 1
-#define remove_sticky(c) c->sticky = 0
-#define toggle_sticky(c) c->sticky = !c->sticky
-
-#define discard_enter_events() do { \
-		XEvent dummy; \
-		XSync(dpy, False); \
-		while (XCheckMaskEvent(dpy, EnterWindowMask, &dummy)); \
-	} while (0)
+#define is_fixed(c) (c->vdesk == VDESK_FIXED)
+#define add_fixed(c) c->vdesk = VDESK_FIXED
+#define remove_fixed(c) c->vdesk = c->screen->vdesk
 
 /* screen structure */
 
@@ -111,13 +116,16 @@ typedef struct ScreenInfo ScreenInfo;
 struct ScreenInfo {
 	int screen;
 	Window root;
+	Window supporting;  /* Dummy window for EWMH */
 	GC invert_gc;
 	XColor fg, bg;
 #ifdef VWM
-	int vdesk;
+	unsigned int vdesk;
 	XColor fc;
+	unsigned old_vdesk; /* most recently unmapped vdesk, so user may toggle back to it */
 #endif
 	char *display;
+	int docks_visible;
 };
 
 /* client structure */
@@ -127,9 +135,7 @@ struct Client {
 	Window  window;
 	Window  parent;
 	ScreenInfo      *screen;
-#ifdef COLOURMAP
 	Colormap        cmap;
-#endif
 	int             ignore_unmap;
 
 	int             x, y, width, height;
@@ -140,14 +146,14 @@ struct Client {
 	int             max_width, max_height;
 	int             width_inc, height_inc;
 	int             base_width, base_height;
+	int             win_gravity_hint;
 	int             win_gravity;
 	int             old_border;
 #ifdef VWM
-	int             vdesk;
-	int             sticky;
+	unsigned int vdesk;
 #endif
+	int             is_dock;
 	int             remove;  /* set when client needs to be removed */
-	Client  *next;
 };
 
 typedef struct Application Application;
@@ -157,11 +163,10 @@ struct Application {
 	int geometry_mask;
 	int x, y;
 	unsigned int width, height;
+	int is_dock;
 #ifdef VWM
-	int vdesk;
-	int sticky;
+	unsigned int vdesk;
 #endif
-	Application *next;
 };
 
 /* Declarations for global variables in main.c */
@@ -185,54 +190,85 @@ extern Atom xa_wm_state;
 extern Atom xa_wm_protos;
 extern Atom xa_wm_delete;
 extern Atom xa_wm_cmapwins;
+
 /* Motif atoms */
 extern Atom mwm_hints;
-/* EWMH atoms */
+
+/* evilwm atoms */
+extern Atom xa_evilwm_unmaximised_horz;
+extern Atom xa_evilwm_unmaximised_vert;
+
+/* EWMH: Root Window Properties (and Related Messages) */
+#ifdef VWM
+extern Atom xa_net_current_desktop;
+#endif
+extern Atom xa_net_active_window;
+
+/* EWMH: Other Root Window Messages */
+extern Atom xa_net_close_window;
+extern Atom xa_net_moveresize_window;
+extern Atom xa_net_restack_window;
+extern Atom xa_net_request_frame_extents;
+
+/* EWMH: Application Window Properties */
 #ifdef VWM
 extern Atom xa_net_wm_desktop;
-extern Atom xa_net_wm_state;
-extern Atom xa_net_wm_state_sticky;
 #endif
+extern Atom xa_net_wm_window_type;
+extern Atom xa_net_wm_window_type_dock;
+extern Atom xa_net_wm_state;
+extern Atom xa_net_wm_state_maximized_vert;
+extern Atom xa_net_wm_state_maximized_horz;
+extern Atom xa_net_wm_state_fullscreen;
+extern Atom xa_net_frame_extents;
 
 /* Things that affect user interaction */
 extern unsigned int     numlockmask;
 extern unsigned int     grabmask1;
 extern unsigned int     grabmask2;
 extern unsigned int     altmask;
-extern const char       *opt_term[3];
+extern char             **opt_term;
 extern int              opt_bw;
-#ifdef SNAP
 extern int              opt_snap;
-#endif
 #ifdef SOLIDDRAG
-extern int              solid_drag;
+extern int              no_solid_drag;
 #else
-# define solid_drag (0)
+# define no_solid_drag (1)
 #endif
-extern Application      *head_app;
+extern struct list      *applications;
 
 /* Client tracking information */
-extern Client           *head_client;
+extern struct list      *clients_tab_order;
+extern struct list      *clients_mapping_order;
+extern struct list      *clients_stacking_order;
 extern Client           *current;
 extern volatile Window  initialising;
 
+/* Event loop will run until this flag is set */
+extern int wm_exit;
+
 /* client.c */
 
 Client *find_client(Window w);
-void gravitate_client(Client *c, int sign);
+void client_hide(Client *c);
+void client_show(Client *c);
+void client_raise(Client *c);
+void client_lower(Client *c);
+void gravitate_border(Client *c, int bw);
 void select_client(Client *c);
 #ifdef VWM
-void fix_client(Client *c);
+void client_to_vdesk(Client *c, unsigned int vdesk);
 #endif
 void remove_client(Client *c);
 void send_config(Client *c);
 void send_wm_delete(Client *c, int kill_client);
 void set_wm_state(Client *c, int state);
 void set_shape(Client *c);
+void *get_property(Window w, Atom property, Atom req_type, unsigned long *nitems_return);
 
 /* events.c */
 
-void event_main_loop(void) __attribute__((noreturn));
+void event_main_loop(void);
 
 /* misc.c */
 
@@ -241,32 +277,48 @@ extern int ignore_xerror;
 int handle_xerror(Display *dsply, XErrorEvent *e);
 void spawn(const char *const cmd[]);
 void handle_signal(int signo);
+void discard_enter_events(Client *except);
 
 /* new.c */
 
 void make_new_client(Window w, ScreenInfo *s);
 long get_wm_normal_hints(Client *c);
+void get_window_type(Client *c);
 
 /* screen.c */
 
 void drag(Client *c);
 void moveresize(Client *c);
-void maximise_client(Client *c, int hv);
-void show_info(Client *c, KeySym key);
+void maximise_client(Client *c, int action, int hv);
+void show_info(Client *c, unsigned int keycode);
 void sweep(Client *c);
-void unhide(Client *c, int raise_win);
 void next(void);
 #ifdef VWM
-void hide(Client *c);
-void switch_vdesk(ScreenInfo *s, int v);
+void switch_vdesk(ScreenInfo *s, unsigned int v);
 #endif
+void set_docks_visible(ScreenInfo *s, int is_visible);
 ScreenInfo *find_screen(Window root);
 ScreenInfo *find_current_screen(void);
 void grab_keys_for_screen(ScreenInfo *s);
 
 /* ewmh.c */
 
+void ewmh_init(void);
+void ewmh_init_screen(ScreenInfo *s);
+void ewmh_deinit_screen(ScreenInfo *s);
+void ewmh_init_client(Client *c);
+void ewmh_deinit_client(Client *c);
+void ewmh_withdraw_client(Client *c);
+void ewmh_select_client(Client *c);
+void ewmh_set_net_client_list(ScreenInfo *s);
+void ewmh_set_net_client_list_stacking(ScreenInfo *s);
 #ifdef VWM
-void update_net_wm_desktop(Client *c);
-void update_net_wm_state(Client *c);
+void ewmh_set_net_current_desktop(ScreenInfo *s);
 #endif
+void ewmh_set_net_active_window(Client *c);
+#ifdef VWM
+void ewmh_set_net_wm_desktop(Client *c);
+#endif
+unsigned int ewmh_get_net_wm_window_type(Window w);
+void ewmh_set_net_wm_state(Client *c);
+void ewmh_set_net_frame_extents(Window w);
