@@ -1,5 +1,5 @@
 /* evilwm - Minimalist Window Manager for X
- * Copyright (C) 1999-2007 Ciaran Anscomb <evilwm@6809.org.uk>
+ * Copyright (C) 1999-2011 Ciaran Anscomb <evilwm@6809.org.uk>
  * see README for license and other details. */
 
 #include <stdlib.h>
@@ -9,6 +9,11 @@
 #include <X11/cursorfont.h>
 #include "evilwm.h"
 #include "log.h"
+#include "xconfig.h"
+
+#ifdef DEBUG
+int log_indent = 0;
+#endif
 
 /* Commonly used X information */
 Display     *dpy;
@@ -24,21 +29,8 @@ int         have_shape, shape_event;
 int         have_randr, randr_event_base;
 #endif
 
-/* Standard X protocol atoms */
-Atom xa_wm_state;
-Atom xa_wm_protos;
-Atom xa_wm_delete;
-Atom xa_wm_cmapwins;
-/* Motif atoms */
-Atom mwm_hints;
-/* EWMH atoms */
-#ifdef VWM
-Atom xa_net_wm_desktop;
-Atom xa_net_wm_state;
-Atom xa_net_wm_state_sticky;
-#endif
-
 /* Things that affect user interaction */
+#define CONFIG_FILE ".evilwmrc"
 static const char   *opt_display = "";
 static const char   *opt_font = DEF_FONT;
 static const char   *opt_fg = DEF_FG;
@@ -46,157 +38,140 @@ static const char   *opt_bg = DEF_BG;
 #ifdef VWM
 static const char   *opt_fc = DEF_FC;
 #endif
+static char *opt_grabmask1 = NULL;
+static char *opt_grabmask2 = NULL;
+static char *opt_altmask = NULL;
 unsigned int numlockmask = 0;
 unsigned int grabmask1 = ControlMask|Mod1Mask;
 unsigned int grabmask2 = Mod1Mask;
 unsigned int altmask = ShiftMask;
-const char   *opt_term[3] = { DEF_TERM, DEF_TERM, NULL };
+static const char *const def_term[] = { DEF_TERM, NULL };
+char **opt_term = (char **)def_term;
 int          opt_bw = DEF_BW;
-#ifdef SNAP
 int          opt_snap = 0;
-#endif
 #ifdef SOLIDDRAG
-int          solid_drag = 1;  /* use solid drag by default */
+int          no_solid_drag = 0;  /* use solid drag by default */
 #endif
-Application  *head_app = NULL;
+struct list  *applications = NULL;
 
 /* Client tracking information */
-Client          *head_client = NULL;
+struct list     *clients_tab_order = NULL;
+struct list     *clients_mapping_order = NULL;
+struct list     *clients_stacking_order = NULL;
 Client          *current = NULL;
 volatile Window initialising = None;
 
-/* Simple command-line options */
-static struct simple_option {
-	const char *opt;
-	enum { OPT_INT, OPT_STRING, OPT_MODIFIERS } type;
-	void *dest;
-} option_list[] = {
-	{ "-fn", OPT_STRING, &opt_font },
-	{ "-display", OPT_STRING, &opt_display },
-	{ "-fg", OPT_STRING, &opt_fg },
-	{ "-bg", OPT_STRING, &opt_bg },
-#ifdef VWM
-	{ "-fc", OPT_STRING, &opt_fc },
-#endif
-	{ "-bw", OPT_INT, &opt_bw },
-	{ "-term", OPT_STRING, &opt_term[0] },
-#ifdef SNAP
-	{ "-snap", OPT_INT, &opt_snap },
-#endif
-	{ "-mask1", OPT_MODIFIERS, &grabmask1 },
-	{ "-mask2", OPT_MODIFIERS, &grabmask2 },
-	{ "-altmask", OPT_MODIFIERS, &altmask },
-};
+/* Event loop will run until this flag is set */
+int wm_exit;
 
-#define NUM_SIMPLE_OPTS (int)(sizeof(option_list) / sizeof(struct simple_option))
+static void set_app(const char *arg);
+static void set_app_geometry(const char *arg);
+static void set_app_dock(void);
+#ifdef VWM
+static void set_app_vdesk(const char *arg);
+static void set_app_fixed(void);
+#endif
+
+static struct xconfig_option evilwm_options[] = {
+	{ XCONFIG_STRING,   "fn",           &opt_font },
+	{ XCONFIG_STRING,   "display",      &opt_display },
+	{ XCONFIG_STRING,   "fg",           &opt_fg },
+	{ XCONFIG_STRING,   "bg",           &opt_bg },
+#ifdef VWM
+	{ XCONFIG_STRING,   "fc",           &opt_fc },
+#endif
+	{ XCONFIG_INT,      "bw",           &opt_bw },
+	{ XCONFIG_STR_LIST, "term",         &opt_term },
+	{ XCONFIG_INT,      "snap",         &opt_snap },
+	{ XCONFIG_STRING,   "mask1",        &opt_grabmask1 },
+	{ XCONFIG_STRING,   "mask2",        &opt_grabmask2 },
+	{ XCONFIG_STRING,   "altmask",      &opt_altmask },
+	{ XCONFIG_CALL_1,   "app",          &set_app },
+	{ XCONFIG_CALL_1,   "geometry",     &set_app_geometry },
+	{ XCONFIG_CALL_1,   "g",            &set_app_geometry },
+	{ XCONFIG_CALL_0,   "dock",         &set_app_dock },
+#ifdef VWM
+	{ XCONFIG_CALL_1,   "vdesk",        &set_app_vdesk },
+	{ XCONFIG_CALL_1,   "v",            &set_app_vdesk },
+	{ XCONFIG_CALL_0,   "fixed",        &set_app_fixed },
+	{ XCONFIG_CALL_0,   "f",            &set_app_fixed },
+	{ XCONFIG_CALL_0,   "s",            &set_app_fixed },
+#endif
+#ifdef SOLIDDRAG
+	{ XCONFIG_BOOL,     "nosoliddrag",  &no_solid_drag },
+#endif
+	{ XCONFIG_END, NULL, NULL }
+};
 
 static void setup_display(void);
 static void *xmalloc(size_t size);
 static unsigned int parse_modifiers(char *s);
 
-int main(int argc, char *argv[]) {
-	struct sigaction act;
-	int i;
-
-	for (i = 1; i < argc; i++) {
-		if (i+1 < argc) {
-			int done = 0;
-			int j;
-			for (j = 0; j < NUM_SIMPLE_OPTS; j++) {
-				if (0 == strcmp(argv[i], option_list[j].opt)) {
-					i++;
-					switch (option_list[j].type) {
-					case OPT_STRING:
-						*((char **)option_list[j].dest) = argv[i];
-						break;
-					default:
-					case OPT_INT:
-						*((int *)option_list[j].dest) = atoi(argv[i]);
-						break;
-					case OPT_MODIFIERS:
-						*((unsigned int *)option_list[j].dest) = parse_modifiers(argv[i]);
-						break;
-					}
-					done = 1;
-					break;
-				}
-			}
-			if (done)
-				continue;
-		}
-
-		if (!strcmp(argv[i], "-app") && i+1<argc) {
-			Application *new = xmalloc(sizeof(Application));
-			char *tmp;
-			i++;
-			new->res_name = new->res_class = NULL;
-			new->geometry_mask = 0;
+#ifdef STDIO
+static void helptext(void) {
+	puts(
+"usage: evilwm [-display display] [-term termprog] [-fn fontname]\n"
+"              [-fg foreground]"
 #ifdef VWM
-			new->vdesk = -1;
-			new->sticky = 0;
+" [-fc fixed]"
 #endif
-			if ((tmp = strchr(argv[i], '/'))) {
-				*(tmp++) = 0;
-			}
-			if (strlen(argv[i]) > 0) {
-				new->res_name = xmalloc(strlen(argv[i])+1);
-				strcpy(new->res_name, argv[i]);
-			}
-			if (tmp && strlen(tmp) > 0) {
-				new->res_class = xmalloc(strlen(tmp)+1);
-				strcpy(new->res_class, tmp);
-			}
-			new->next = head_app;
-			head_app = new;
-		} else if (!strcmp(argv[i], "-g") && i+1<argc) {
-			i++;
-			if (!head_app)
-				continue;
-			head_app->geometry_mask = XParseGeometry(argv[i],
-					&head_app->x, &head_app->y,
-					&head_app->width, &head_app->height);
+" [-bg background] [-bw borderwidth]\n"
+"              [-mask1 modifiers] [-mask2 modifiers] [-altmask modifiers]\n"
+"              [-snap num]"
+" [-app name/class] [-g geometry] [-dock]\n"
 #ifdef VWM
-		} else if (!strcmp(argv[i], "-v") && i+1<argc) {
-			int v = atoi(argv[++i]);
-			if (head_app && valid_vdesk(v))
-				head_app->vdesk = v;
-		} else if (!strcmp(argv[i], "-s")) {
-			if (head_app)
-				head_app->sticky = 1;
+"              [-v vdesk] [-s]"
 #endif
 #ifdef SOLIDDRAG
-		} else if (!strcmp(argv[i], "-nosoliddrag")) {
-			solid_drag = 0;
+" [-nosoliddrag]"
 #endif
+" [-V]"
+	);
+}
+#else
+#define helptext()
+#endif
+
+int main(int argc, char *argv[]) {
+	struct sigaction act;
+	int argn = 1, ret;
+
+	{
+		const char *home = getenv("HOME");
+		if (home) {
+			char *conffile = xmalloc(strlen(home) + sizeof(CONFIG_FILE) + 2);
+			strcpy(conffile, home);
+			strcat(conffile, "/" CONFIG_FILE);
+			xconfig_parse_file(evilwm_options, conffile);
+			free(conffile);
+		}
+	}
+	ret = xconfig_parse_cli(evilwm_options, argc, argv, &argn);
+	if (ret == XCONFIG_MISSING_ARG) {
+		fprintf(stderr, "%s: missing argument to `%s'\n", argv[0], argv[argn]);
+		exit(1);
+	} else if (ret == XCONFIG_BAD_OPTION) {
+		if (0 == strcmp(argv[argn], "-h")
+				|| 0 == strcmp(argv[argn], "--help")) {
+			helptext();
+			exit(0);
 #ifdef STDIO
-		} else if (!strcmp(argv[i], "-V")) {
+		} else if (0 == strcmp(argv[argn], "-V")
+				|| 0 == strcmp(argv[argn], "--version")) {
 			LOG_INFO("evilwm version " VERSION "\n");
 			exit(0);
 #endif
 		} else {
-			LOG_INFO("usage: evilwm [-display display] [-term termprog] [-fn fontname]\n");
-			LOG_INFO("              [-fg foreground]");
-#ifdef VWM
-			LOG_INFO(" [-fc fixed]");
-#endif
-			LOG_INFO(" [-bg background] [-bw borderwidth]\n");
-			LOG_INFO("              [-mask1 modifiers] [-mask2 modifiers] [-altmask modifiers]\n");
-			LOG_INFO("              [-snap num]");
-#ifdef VWM
-			LOG_INFO(" [-app name/class] [-g geometry] [-v vdesk] [-s]");
-#endif
-#ifdef SOLIDDRAG
-			LOG_INFO("\n              [-nosoliddrag]");
-#endif
-			LOG_INFO(" [-V]\n");
-			if (0 == strcmp(argv[i], "-h") || 0 == strcmp(argv[i], "--help"))
-				exit(0);
+			helptext();
 			exit(1);
 		}
 	}
 
-	opt_term[1] = opt_term[0];
+	if (opt_grabmask1) grabmask1 = parse_modifiers(opt_grabmask1);
+	if (opt_grabmask2) grabmask2 = parse_modifiers(opt_grabmask2);
+	if (opt_altmask) altmask = parse_modifiers(opt_altmask);
 
+	wm_exit = 0;
 	act.sa_handler = handle_signal;
 	sigemptyset(&act.sa_mask);
 	act.sa_flags = 0;
@@ -208,7 +183,23 @@ int main(int argc, char *argv[]) {
 
 	event_main_loop();
 
-	return 1;
+	/* Quit Nicely */
+	while (clients_stacking_order)
+		remove_client(clients_stacking_order->data);
+	XSetInputFocus(dpy, PointerRoot, RevertToPointerRoot, CurrentTime);
+	if (font) XFreeFont(dpy, font);
+	{
+		int i;
+		for (i = 0; i < num_screens; i++) {
+			ewmh_deinit_screen(&screens[i]);
+			XFreeGC(dpy, screens[i].invert_gc);
+			XInstallColormap(dpy, DefaultColormap(dpy, i));
+		}
+	}
+	free(screens);
+	XCloseDisplay(dpy);
+
+	return 0;
 }
 
 static void *xmalloc(size_t size) {
@@ -237,29 +228,18 @@ static void setup_display(void) {
 	Window dw1, dw2, *wins;
 	XWindowAttributes winattr;
 
+	LOG_ENTER("setup_display()");
+
 	dpy = XOpenDisplay(opt_display);
-	if (!dpy) { 
+	if (!dpy) {
 		LOG_ERROR("can't open display %s\n", opt_display);
 		exit(1);
 	}
 	XSetErrorHandler(handle_xerror);
 	/* XSynchronize(dpy, True); */
 
-	/* Standard X protocol atoms */
-	xa_wm_state = XInternAtom(dpy, "WM_STATE", False);
-	xa_wm_protos = XInternAtom(dpy, "WM_PROTOCOLS", False);
-	xa_wm_delete = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
-#ifdef COLOURMAP
-	xa_wm_cmapwins = XInternAtom(dpy, "WM_COLORMAP_WINDOWS", False);
-#endif
-	/* Motif atoms */
-	mwm_hints = XInternAtom(dpy, _XA_MWM_HINTS, False);
-	/* EWMH atoms */
-#ifdef VWM
-	xa_net_wm_desktop = XInternAtom(dpy, "_NET_WM_DESKTOP", False);
-	xa_net_wm_state = XInternAtom(dpy, "_NET_WM_STATE", False);
-	xa_net_wm_state_sticky = XInternAtom(dpy, "_NET_WM_STATE_STICKY", False);
-#endif
+	/* Standard & EWMH atoms */
+	ewmh_init();
 
 	font = XLoadQueryFont(dpy, opt_font);
 	if (!font) font = XLoadQueryFont(dpy, DEF_FONT);
@@ -278,7 +258,7 @@ static void setup_display(void) {
 		for (j = 0; j < (unsigned int)modmap->max_keypermod; j++) {
 			if (modmap->modifiermap[i*modmap->max_keypermod+j] == XKeysymToKeycode(dpy, XK_Num_Lock)) {
 				numlockmask = (1<<i);
-				LOG_DEBUG("setup_display() : XK_Num_Lock is (1<<0x%02x)\n", i);
+				LOG_DEBUG("XK_Num_Lock is (1<<0x%02x)\n", i);
 			}
 		}
 	}
@@ -291,11 +271,7 @@ static void setup_display(void) {
 	gv.font = font->fid;
 
 	/* set up root window attributes - same for each screen */
-#ifdef COLOURMAP
 	attr.event_mask = ChildMask | EnterWindowMask | ColormapChangeMask;
-#else
-	attr.event_mask = ChildMask | EnterWindowMask;
-#endif
 
 	/* SHAPE extension? */
 #ifdef SHAPE
@@ -309,10 +285,9 @@ static void setup_display(void) {
 	{
 		int e_dummy;
 		have_randr = XRRQueryExtension(dpy, &randr_event_base, &e_dummy);
-#ifdef DEBUG
-		if (!have_randr)
-			LOG_DEBUG("Xrandr is not supported on this display.\n");
-#endif
+		if (!have_randr) {
+			LOG_DEBUG("XRandR is not supported on this display.\n");
+		}
 	}
 #endif
 
@@ -328,17 +303,16 @@ static void setup_display(void) {
 		ds = DisplayString(dpy);
 		/* set up DISPLAY environment variable to use */
 		colon = strrchr(ds, ':');
+		screens[i].display = xmalloc(14 + strlen(ds));
+		strcpy(screens[i].display, "DISPLAY=");
+		strcat(screens[i].display, ds);
 		if (colon && num_screens > 1) {
-			screens[i].display = xmalloc(14 + strlen(ds));
-			strcpy(screens[i].display, "DISPLAY=");
-			strcat(screens[i].display, ds);
 			colon = strrchr(screens[i].display, ':');
 			dot = strchr(colon, '.');
 			if (!dot)
 				dot = colon + strlen(colon);
 			snprintf(dot, 5, ".%d", i);
-		} else
-			screens[i].display = NULL;
+		}
 
 		screens[i].screen = i;
 		screens[i].root = RootWindow(dpy, i);
@@ -361,20 +335,83 @@ static void setup_display(void) {
 
 		XChangeWindowAttributes(dpy, screens[i].root, CWEventMask, &attr);
 		grab_keys_for_screen(&screens[i]);
+		screens[i].docks_visible = 1;
 
 		/* scan all the windows on this screen */
-		LOG_XDEBUG("main:XQueryTree(); ");
+		LOG_XENTER("XQueryTree(screen=%d)", i);
 		XQueryTree(dpy, screens[i].root, &dw1, &dw2, &wins, &nwins);
 		LOG_XDEBUG("%d windows\n", nwins);
+		LOG_XLEAVE();
 		for (j = 0; j < nwins; j++) {
 			XGetWindowAttributes(dpy, wins[j], &winattr);
 			if (!winattr.override_redirect && winattr.map_state == IsViewable)
 				make_new_client(wins[j], &screens[i]);
 		}
 		XFree(wins);
+		ewmh_init_screen(&screens[i]);
+	}
+	ewmh_set_net_active_window(NULL);
+	LOG_LEAVE();
+}
+
+/**************************************************************************/
+/* Option parsing callbacks */
+
+static void set_app(const char *arg) {
+	Application *new = xmalloc(sizeof(Application));
+	char *tmp;
+	new->res_name = new->res_class = NULL;
+	new->geometry_mask = 0;
+	new->is_dock = 0;
+#ifdef VWM
+	new->vdesk = VDESK_NONE;
+#endif
+	if ((tmp = strchr(arg, '/'))) {
+		*(tmp++) = 0;
+	}
+	if (strlen(arg) > 0) {
+		new->res_name = xmalloc(strlen(arg)+1);
+		strcpy(new->res_name, arg);
+	}
+	if (tmp && strlen(tmp) > 0) {
+		new->res_class = xmalloc(strlen(tmp)+1);
+		strcpy(new->res_class, tmp);
+	}
+	applications = list_prepend(applications, new);
+}
+
+static void set_app_geometry(const char *arg) {
+	if (applications) {
+		Application *app = applications->data;
+		app->geometry_mask = XParseGeometry(arg,
+				&app->x, &app->y, &app->width, &app->height);
 	}
 }
 
+static void set_app_dock(void) {
+	if (applications) {
+		Application *app = applications->data;
+		app->is_dock = 1;
+	}
+}
+
+#ifdef VWM
+static void set_app_vdesk(const char *arg) {
+	unsigned int v = atoi(arg);
+	if (applications && valid_vdesk(v)) {
+		Application *app = applications->data;
+		app->vdesk = v;
+	}
+}
+
+static void set_app_fixed(void) {
+	if (applications) {
+		Application *app = applications->data;
+		app->vdesk = VDESK_FIXED;
+	}
+}
+#endif
+
 /* Used for overriding the default WM modifiers */
 static unsigned int parse_modifiers(char *s) {
 	static struct {
