@@ -1,11 +1,12 @@
 /* evilwm - Minimalist Window Manager for X
- * Copyright (C) 1999-2007 Ciaran Anscomb <evilwm@6809.org.uk>
+ * Copyright (C) 1999-2011 Ciaran Anscomb <evilwm@6809.org.uk>
  * see README for license and other details. */
 
 #include <X11/Xproto.h>
 #include <stdlib.h>
 #include <stdarg.h>
 #include <signal.h>
+#include <string.h>
 #include <sys/wait.h>
 #include <unistd.h>
 #include "evilwm.h"
@@ -24,12 +25,12 @@ void spawn(const char *const cmd[]) {
 	if (!(pid = fork())) {
 		setsid();
 		switch (fork()) {
-			/* Expect compiler warnings because of half-broken SUS
-			 * execvp prototype:  "char *const argv[]" should have
-			 * been "const char *const argv[]", but the committee
-			 * favored legacy code over modern code, and modern
-			 * compilers bark at our extra const. (LD) */
-			case 0: execvp(cmd[0], cmd+1);
+			/* execvp()'s prototype is (char *const *) suggesting that it
+			 * modifies the contents of the strings.  The prototype is this
+			 * way due to SUS maintaining compatability with older code.
+			 * However, execvp guarantees not to modify argv, so the following
+			 * cast is valid. */
+			case 0: execvp(cmd[0], (char *const *)cmd);
 			default: _exit(0);
 		}
 	}
@@ -38,39 +39,29 @@ void spawn(const char *const cmd[]) {
 }
 
 void handle_signal(int signo) {
-	int i;
 	(void)signo;  /* unused */
-	/* SIGCHLD check no longer necessary */
-	/* Quit Nicely */
-	while(head_client) remove_client(head_client);
-	XSetInputFocus(dpy, PointerRoot, RevertToPointerRoot, CurrentTime);
-	if (font) XFreeFont(dpy, font);
-	for (i = 0; i < num_screens; i++) {
-		XFreeGC(dpy, screens[i].invert_gc);
-		XInstallColormap(dpy, DefaultColormap(dpy, i));
-	}
-	free(screens);
-	XCloseDisplay(dpy);
-	exit(0);
+	wm_exit = 1;
 }
 
 int handle_xerror(Display *dsply, XErrorEvent *e) {
 	Client *c;
 	(void)dsply;  /* unused */
 
+	LOG_ENTER("handle_xerror(error=%d, request=%d/%d, resourceid=%lx)", e->error_code, e->request_code, e->minor_code, e->resourceid);
+
 	if (ignore_xerror) {
-		LOG_DEBUG("handle_xerror() ignored an XErrorEvent: %d\n", e->error_code);
+		LOG_DEBUG("ignoring...\n");
+		LOG_LEAVE();
 		return 0;
 	}
 	/* If this error actually occurred while setting up the new
 	 * window, best let make_new_client() know not to bother */
 	if (initialising != None && e->resourceid == initialising) {
-		LOG_DEBUG("\t **SAVED?** handle_xerror() caught error %d while initialising\n", e->error_code);
+		LOG_DEBUG("error caught while initialising window=%lx\n", initialising);
 		initialising = None;
+		LOG_LEAVE();
 		return 0;
 	}
-	LOG_DEBUG("**ERK** handle_xerror() caught an XErrorEvent: error_code=%d request_code=%d minor_code=%d\n",
-			e->error_code, e->request_code, e->minor_code);
 	if (e->error_code == BadAccess && e->request_code == X_ChangeWindowAttributes) {
 		LOG_ERROR("root window unavailable (maybe another wm is running?)\n");
 		exit(1);
@@ -78,9 +69,29 @@ int handle_xerror(Display *dsply, XErrorEvent *e) {
 
 	c = find_client(e->resourceid);
 	if (c) {
-		LOG_DEBUG("\thandle_xerror() : flagging client for removal\n");
+		LOG_DEBUG("flagging client for removal\n");
 		c->remove = 1;
 		need_client_tidy = 1;
+	} else {
+		LOG_DEBUG("unknown error: not handling\n");
 	}
+	LOG_LEAVE();
 	return 0;
 }
+
+/* Remove all enter events from the queue except the last of any corresponding
+ * to "except"s parent. */
+void discard_enter_events(Client *except) {
+	XEvent tmp, putback_ev;
+	int putback = 0;
+	XSync(dpy, False);
+	while (XCheckMaskEvent(dpy, EnterWindowMask, &tmp)) {
+		if (tmp.xcrossing.window == except->parent) {
+			memcpy(&putback_ev, &tmp, sizeof(XEvent));
+			putback = 1;
+		}
+	}
+	if (putback) {
+		XPutBackEvent(dpy, &putback_ev);
+	}
+}
